@@ -11,7 +11,9 @@
 -- Functions for human-readable derived 'Show' instances.
 --------------------------------------------------------------------------------
 
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE Safe #-}
+{-# LANGUAGE PatternGuards #-}
 module Text.Show.Pretty
   ( -- * Generic representation of values
     Value(..), Name
@@ -26,7 +28,7 @@ module Text.Show.Pretty
     ppDocList, ppShowList, pPrintList
 
     -- * Values using the 'PrettyVal' class
-  , dumpDoc, dumpStr, PrettyVal(..)
+  , dumpDoc, dumpStr, dumpIO, PrettyVal(..)
 
     -- * Rendering values to Html
   , valToHtml, HtmlOpts(..), defaultHtmlOpts, htmlPage, Html(..)
@@ -34,19 +36,29 @@ module Text.Show.Pretty
     -- * Get location of data files
   , getDataDir
 
+  , -- * Preprocessing of values
+    PreProc(..), ppHide, ppHideNested, hideCon
+
     -- * Deprecated
   , ppValue
   ) where
 
+import Data.Char(isHexDigit)
 import Text.PrettyPrint
 import qualified Text.Show.Parser as P
 import Text.Show.Value
 import Text.Show.PrettyVal
 import Text.Show.Html
 import Data.Foldable(Foldable,toList)
-import Language.Haskell.Lexer(rmSpace,lexerPass0)
+import Language.Haskell.Lexer(rmSpace,lexerPass0,Token(..))
 import Paths_pretty_show (getDataDir)
 
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>) )
+#else
+import Prelude
+#endif
+
 {-# DEPRECATED ppValue "Please use `valToDoc` instead." #-}
 ppValue :: Value -> Doc
 ppValue = valToDoc
@@ -55,7 +67,20 @@ reify :: Show a => a -> Maybe Value
 reify = parseValue . show
 
 parseValue :: String -> Maybe Value
-parseValue = P.parseValue . rmSpace . lexerPass0
+parseValue = P.parseValue . rmSpace . foldr joinTokens [] . lexerPass0
+  where
+
+  -- Sometimes we join tokens that are next to each other, with no spaces.
+  -- This improves the printing of some malformed inputs:
+  --   * Hex numbers with no 0x:    "4ab"  instead of "4 ab"
+  joinTokens a@(t1,(p1,s1)) bs =
+    case bs of
+      (_t2,(_,s2)) : more
+        | IntLit <- t1, all isHexDigit s2 -> jn IntLit
+          where jn t = (t,(p1,s1++s2)) : more
+
+      _ -> a : bs
+
 
 -- | Convert a generic value into a pretty 'String', if possible.
 ppShow :: Show a => a -> String
@@ -104,6 +129,10 @@ dumpDoc = valToDoc . prettyVal
 dumpStr :: PrettyVal a => a -> String
 dumpStr = show . dumpDoc
 
+-- | Render a value using the 'PrettyVal' class and show it to standard out.
+dumpIO :: PrettyVal a => a -> IO ()
+dumpIO = putStrLn . dumpStr
+
 
 -- | Pretty print a generic value. Our intention is that the result is
 --   equivalent to the 'Show' instance for the original value, except possibly
@@ -138,6 +167,31 @@ valToDoc val = case val of
   String x         -> text x
 
 
+-- | This type is used to allow pre-processing of values before showing them.
+data PreProc a = PreProc (Value -> Value) a
+
+instance Show a => Show (PreProc a) where
+  showsPrec p (PreProc f a) cs =
+    case parseValue txt of
+      Nothing -> txt ++ cs
+      Just v  -> wrap (valToStr (f v))
+    where
+    txt    = showsPrec p a ""
+    wrap t = case (t,txt) of
+              (h:_,'(':_) | h /= '(' -> '(' : (t ++ ')' : cs)
+              _ -> t ++ cs
+
+-- | Hide the given constructors when showing a value.
+ppHide :: (Name -> Bool) -> a -> PreProc a
+ppHide p = PreProc (hideCon False p)
+
+-- | Hide the given constructors when showing a value.
+-- In addition, hide values if all of their children were hidden.
+ppHideNested :: (Name -> Bool) -> a -> PreProc a
+ppHideNested p = PreProc (hideCon True p)
+
+
+
 -- Private ---------------------------------------------------------------------
 
 ppAtom :: Value -> Doc
@@ -177,3 +231,4 @@ blockWith f a b (d:ds)  = f $
     (char a <+> d) : [ char ',' <+> x | x <- ds ] ++ [ char b ]
 
 
+
